from .__meta__ import version as __version__


from dataclasses import Field, field, MISSING, FrozenInstanceError, InitVar, fields, asdict, astuple, make_dataclass, \
    replace, is_dataclass
from .internals import dataclass, field_property
from .datetime_utils import Date, Time, DateTime, TimeDelta, \
    datetime_property, time_property, date_property, timedelta_helper_property, seconds_property
from .weekdays_list import Weekdays, weekdays_property


__all__ = ['dataclass', 'field_property',
           'field',
           'Field',
           'FrozenInstanceError',
           'InitVar',
           'MISSING',

           # Helper functions.
           'fields',
           'asdict',
           'astuple',
           'make_dataclass',
           'replace',
           'is_dataclass',

           # datetime
           'Date', 'Time', 'DateTime', 'TimeDelta',
           'datetime_property', 'time_property', 'date_property', 'timedelta_helper_property', 'seconds_property',

           # Weekdays
           'Weekdays', 'weekdays_property',
           ]



try:
    from dataclass_property import field_property
except (ImportError, Exception):
    class field_property(property):
        def __init__(self, fget=None, fset=None, fdel=None, doc=None, default=None, default_factory=None):
            super().__init__(fget, fset, fdel, doc)


__all__ = ['Weekdays', 'weekdays_property']


def is_iterable(obj):
    """Return if iterable and is not a string."""
    if isinstance(obj, str):
        return False

    try:
        iter(obj)
        return True
    except (TypeError, ValueError, Exception):
        return False


def weekdays_property(attr):
    """Return a property to access Weekdays.sunday as a property.

    Example:
        .. code-block:: python

            class MyClass:
                weekdays: Weekdays = Weekdays()
                sunday: bool = weekdays_property('sunday')
                monday: bool = weekdays_property('monday')
                tuesday: bool = weekdays_property('tuesday')
                wednesday: bool = weekdays_property('wednesday')
                thursday: bool = weekdays_property('thursday')
                friday: bool = weekdays_property('friday')
                saturday: bool = weekdays_property('saturday')
    """
    def fget(self):
        return getattr(self.weekdays, attr, 0)

    def fset(self, value):
        if value is None:
            return  # Ignore initial value of None
        setattr(self.weekdays, attr, value)

    def fdel(self):
        delattr(self.weekdays, attr)

    return field_property(fget, fset, fdel, doc='Alias for weekdays {}'.format(attr), default=None)


def _weekday_prop(name):
    def fget(self):
        return name in self

    def fset(self, value):
        if value:
            self.append(name)
        elif name in self:
            self.remove(name)

    return property(fget, fset)


class Weekdays(list):
    """Sorted list of weekdays.

    Features:
      * If no arguments are given the list will be populated with all weekday names.
      * If `Weekdays(friday=False)` all weekdays will be in the list except for fridays.
      * Weekdays can be added or removed by a property `my_week.monday = False` or `my_week.tuesday = True`.
      * Abbreviations ("Sun", "Mon", "Tue") and weekday indexes (where monday is 0, sunday is 6) are allowed.
        * Abbreviations and indexes work with init, append, insert, extend, remove, __add__, __contains__

    Args:
        *args (tuple/object): Positional arguments of names to be in the list of weekdays.
            ("sunday", "monday", "Tue", "Fri")
        **kwargs (dict/object): Keyword arguments for weekday values.
            {"Sunday": True, "Monday": True, "Tue": True, "Fri": False}
    """
    SUNDAY_VALUE = 'sunday'
    MONDAY_VALUE = 'monday'
    TUESDAY_VALUE = 'tuesday'
    WEDNESDAY_VALUE = 'wednesday'
    THURSDAY_VALUE = 'thursday'
    FRIDAY_VALUE = 'friday'
    SATURDAY_VALUE = 'saturday'

    # Weekday names and order
    DAYS = {SUNDAY_VALUE: 0, MONDAY_VALUE: 1, TUESDAY_VALUE: 2, WEDNESDAY_VALUE: 3,
            THURSDAY_VALUE: 4, FRIDAY_VALUE: 5, SATURDAY_VALUE: 6}

    _all_true = {name: True for name in DAYS}
    _all_false = {name: False for name in DAYS}

    MAPPING = {'sunday': SUNDAY_VALUE, 'sundays': SUNDAY_VALUE, 'sun': SUNDAY_VALUE, 6: SUNDAY_VALUE,
               'monday': MONDAY_VALUE, 'mondays': MONDAY_VALUE, 'mon': MONDAY_VALUE, 0: MONDAY_VALUE,
               'tuesday': TUESDAY_VALUE, 'tuesdays': TUESDAY_VALUE, 'tue': TUESDAY_VALUE, 1: TUESDAY_VALUE,
               'wednesday': WEDNESDAY_VALUE, 'wednesdays': WEDNESDAY_VALUE, 'wed': WEDNESDAY_VALUE, 2: WEDNESDAY_VALUE,
               'thursday': THURSDAY_VALUE, 'thursdays': THURSDAY_VALUE, 'thu': THURSDAY_VALUE, 3: THURSDAY_VALUE,
               'friday': FRIDAY_VALUE, 'fridays': FRIDAY_VALUE, 'fri': FRIDAY_VALUE, 4: FRIDAY_VALUE,
               'saturday': SATURDAY_VALUE, 'saturdays': SATURDAY_VALUE, 'sat': SATURDAY_VALUE, 5: SATURDAY_VALUE,
               }

    sunday = _weekday_prop('sunday')
    monday = _weekday_prop('monday')
    tuesday = _weekday_prop('tuesday')
    wednesday = _weekday_prop('wednesday')
    thursday = _weekday_prop('thursday')
    friday = _weekday_prop('friday')
    saturday = _weekday_prop('saturday')

    sun = _weekday_prop('sunday')
    mon = _weekday_prop('monday')
    tue = _weekday_prop('tuesday')
    wed = _weekday_prop('wednesday')
    thu = _weekday_prop('thursday')
    fri = _weekday_prop('friday')
    sat = _weekday_prop('saturday')

    def __init__(self, *args, **kwargs):
        """Initialize the sorted list of Weekdays

        An empty list means that every day of the week is valid.
        If Weekdays(friday=False) all weekdays will be in the list except for fridays.

        Args:
            *args (tuple/object): Positional arguments of names to be in the list of weekdays.
                ("sunday", "monday", "Tue", "fri")
            **kwargs (dict/object): Keyword arguments for weekday values.
                {"Sunday": True, "Monday": True, "Tue": True, "fri": False}
        """
        if len(args) == 0 and len(kwargs) == 0:
            # If no arguments given assume all True
            values = self._all_true.copy()
        else:
            if len(args) == 1 and is_iterable(args[0]):
                args = args[0]
            values = self._all_false.copy()
            values.update({self.as_attr(arg): True for arg in args if self.is_attr(arg)})
            values.update({self.as_attr(k): v for k, v in kwargs.items() if self.is_attr(k)})

        super().__init__((k for k, v in values.items() if v))

    @classmethod
    def is_attr(cls, value):
        if isinstance(value, str):
            value = value.lower()
        return value in cls.MAPPING

    @classmethod
    def as_attr(cls, value):
        if isinstance(value, str):
            value = value.lower()
        value = cls.MAPPING.get(value, None)
        if value not in cls.DAYS:
            raise ValueError('Invalid weekday given!')

        return value

    def is_valid(self, weekday):
        """Return if the given weekday is in this list of Weekdays."""
        return weekday in self

    def append(self, value):
        value = self.as_attr(value)
        if not list.__contains__(self, value):
            super().append(value)
            self.sort(key=self.DAYS.get)

    def insert(self, index, value):
        value = self.as_attr(value)
        if not list.__contains__(self, value):
            super().insert(index, value)
            self.sort(key=self.DAYS.get)

    def remove(self, value):
        value = self.as_attr(value)
        if list.__contains__(self, value):
            super().remove(value)

    def __contains__(self, value):
        try:
            value = self.as_attr(value)
            return super().__contains__(value)
        except (TypeError, ValueError, Exception):
            return False

    def __setitem__(self, key, value):
        value = self.as_attr(value)
        super().__setitem__(key, value)
        self.sort(key=self.DAYS.get)

    def extend(self, other):
        for n in other:
            if self.is_attr(n):
                n = self.as_attr(n)
                if n not in self:
                    self.append(n)
        self.sort(key=self.DAYS.get)

    def __add__(self, other):
        if isinstance(other, str):
            other = [other]
        new_obj = self.__class__(*self)
        new_obj.extend(other)
        return new_obj

    def __radd__(self, other):
        if isinstance(other, str):
            other = [other]
        new_obj = self.__class__(*self)
        new_obj.extend(other)
        return new_obj

    def __iadd__(self, other):
        if isinstance(other, str):
            other = [other]
        self.extend(other)
        return self

    # ===== Pydantic validation =====
    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
                options=[cls.SUNDAY_VALUE, cls.MONDAY_VALUE, cls.TUESDAY_VALUE, cls.WEDNESDAY_VALUE,
                         cls.THURSDAY_VALUE, cls.FRIDAY_VALUE, cls.SATURDAY_VALUE],
                )

    @classmethod
    def validate(cls, v):
        if isinstance(v, cls):
            return v
        elif v is None:
            v = []
        elif not isinstance(v, (list, tuple)):
            v = [v]
        elif isinstance(v, dict):
            return cls(**v)
        return cls(*v)



import datetime
from typing import Union, List
from dataclass_property import field_property, MISSING


__all__ = ['Date', 'Time', 'DateTime', 'TimeDelta',
           'datetime_property', 'time_property', 'date_property', 'timedelta_helper_property', 'seconds_property']


def date_property(attr, allow_none=True, default=MISSING, default_factory=MISSING, formats: List[str] = None):
    """Create a date dataclass property where the underlying datetime is saved to "_attr".

    Args:
        attr (str): Attribute name (example: "created_on"
        allow_none (bool)[True]: Allows the property to be set to None. This is needed if the default is None.
        default (object)[MISSING]: Default value for the dataclass
        default_factory (function)[MISSING]: Function that returns the default value.
        formats (list)[None]: List of string formats to accept.

    Returns:
        property (field_property): Dataclass field property for a date.
    """
    attr = '_' + attr
    typeref = Union[Date, str]
    if allow_none:
        typeref = Union[Date, str, None]

    if default == MISSING and default_factory == MISSING and allow_none:
        default = None

    def fget(self):
        return getattr(self, attr)

    def fset(self, value: typeref):
        if value is None and not allow_none:
            raise TypeError('Invalid date value given!')
        elif value is not None:
            value = Date.make_date(value, formats=formats)
        setattr(self, attr, value)

    return field_property(fget, fset, doc='date property {}'.format(attr),
                          default=default, default_factory=default_factory)


def time_property(attr, allow_none=True, default=MISSING, default_factory=MISSING, formats: List[str] = None):
    """Create a time dataclass property where the underlying datetime is saved to "_attr".

    Args:
        attr (str): Attribute name (example: "created_on"
        allow_none (bool)[True]: Allows the property to be set to None. This is needed if the default is None.
        default (object)[MISSING]: Default value for the dataclass
        default_factory (function)[MISSING]: Function that returns the default value.
        formats (list)[None]: List of string formats to accept.

    Returns:
        property (field_property): Dataclass field property for a time.
    """
    attr = '_' + attr
    typeref = Union[Time, str]
    if allow_none:
        typeref = Union[Time, str, None]

    if default == MISSING and default_factory == MISSING and allow_none:
        default = None

    def fget(self):
        return getattr(self, attr)

    def fset(self, value: typeref):
        if value is None and not allow_none:
            raise TypeError('Invalid time value given!')
        elif value is not None:
            value = Time.make_time(value, formats=formats)
        setattr(self, attr, value)

    return field_property(fget, fset, doc='time property {}'.format(attr),
                          default=default, default_factory=default_factory)


def datetime_property(attr, allow_none=True, default=MISSING, default_factory=MISSING, formats: List[str] = None):
    """Create a datetime dataclass property where the underlying datetime is saved to "_attr".

    Args:
        attr (str): Attribute name (example: "created_on"
        allow_none (bool)[True]: Allows the property to be set to None. This is needed if the default is None.
        default (object)[MISSING]: Default value for the dataclass
        default_factory (function)[MISSING]: Function that returns the default value.
        formats (list)[None]: List of string formats to accept.

    Returns:
        property (field_property): Dataclass field property for a datetime.
    """
    attr = '_' + attr
    typeref = Union[DateTime, str]
    if allow_none:
        typeref = Union[DateTime, str, None]

    if default == MISSING and default_factory == MISSING and allow_none:
        default = None

    def fget(self):
        return getattr(self, attr)

    def fset(self, value: typeref):
        if value is None and not allow_none:
            raise TypeError('Invalid datetime value given!')
        elif value is not None:
            value = DateTime.make_datetime(value, formats=formats)
        setattr(self, attr, value)

    return field_property(fget, fset, doc='datetime property {}'.format(attr),
                          default=default, default_factory=default_factory)


def timedelta_helper_property():
    """Return a timedelta property that uses a classes attributes for
    'weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds', and 'microseconds'.

    Example:

        .. code-block:: python

            @dataclass
            class MyClass:
                days: int = 0
                hours: int = 0
                minutes: int = 0
                seconds: int = 0
                milliseconds: int = 0

                interval: datetime.timedelta = timedelta_helper_property()

    Returns:
        property (field_property): Timedelta property that uses a classes underlying attributes.
    """
    def fget(self) -> datetime.timedelta:
        weeks = getattr(self, 'weeks', 0)
        days = getattr(self, 'days', 0)
        hours = getattr(self, 'hours', 0)
        minutes = getattr(self, 'minutes', 0)
        seconds = getattr(self, 'seconds', 0)
        milliseconds = getattr(self, 'milliseconds', 0)
        microseconds = getattr(self, 'microseconds', 0)

        return datetime.timedelta(weeks=weeks, days=days, hours=hours, minutes=minutes, seconds=seconds,
                                  milliseconds=milliseconds, microseconds=microseconds)

    def fset(self, value: datetime.timedelta):
        if value is None:  # Protect the default of None
            return
        elif not isinstance(value, datetime.timedelta):
            raise TypeError('Invalid timedelta given!')

        days = value.days
        seconds = value.seconds
        microseconds = value.microseconds

        weeks = int(days / 7)
        days = int(days - weeks)
        if hasattr(self, 'weeks'):
            self.weeks = weeks
        if hasattr(self, 'days'):
            self.days = days

        hours = int(seconds / 3600)
        minutes = int((seconds - (hours * 3600)) / 60)
        seconds = int(seconds - (minutes * 60))
        if hasattr(self, 'hours'):
            self.hours = hours
        if hasattr(self, 'minutes'):
            self.minutes = minutes
        if hasattr(self, 'seconds'):
            self.seconds = seconds

        milliseconds = int(microseconds / 1000)
        microseconds = int(microseconds - (milliseconds * 1000))
        if hasattr(self, 'milliseconds'):
            self.milliseconds = milliseconds
        if hasattr(self, 'microseconds'):
            self.microseconds = microseconds

    return field_property(fget, fset, doc='Interval from the set time values', default=None)


def seconds_property(attr='seconds'):
    """Property for a seconds attribute to turn floating points into milliseconds.

    Args:
        attr (str)['seconds']: Attribute name

    Returns:
        property (field_property): Property that when given a float will set seconds and milliseconds.
    """
    attr = '_' + attr

    def fget(self) -> int:
        return getattr(self, attr, 0)

    def fset(self, value: Union[int, float]):
        if isinstance(value, float):
            mill = int((value % 1) * 1000)
            if mill:
                self.milliseconds = mill

        setattr(self, attr, int(value))

    return field_property(fget, fset, doc='Seconds in time. If float also set milliseconds', default=0)


class Date(datetime.date):
    """DateTime class for pydantic which converts custom string formats into datetimes."""

    DATE_FORMATS = [
        '%Y-%m-%d', '%m/%d/%Y',   # '2019-04-17', '04/17/2019'
        '%b %d %Y', '%b %d, %Y',  # 'Apr 17 2019', 'Apr 17, 2019'
        '%d %b %Y', '%d %b, %Y',  # '17 Apr 2019', '17 Apr, 2019'
        '%B %d %Y', '%B %d, %Y',  # 'April 17 2019', 'April 17, 2019'
        '%d %B %Y', '%d %B, %Y',  # '17 April 2019', '17 April, 2019'
        ]

    @classmethod
    def make_date(cls, date_string: Union[datetime.date, str], formats: List[str] = None) -> datetime.date:
        """Make the date object from the given time string.

        Args:
            date_string (str): Date string 'mm/dd/yyyy' ...
            formats (list): List of acceptable time string formats.

        Returns:
            d (datetime.date): Date object or None.
        """
        if isinstance(date_string, datetime.date):
            return date_string

        if formats is None:
            formats = [i for i in cls.DATE_FORMATS]

        for fmt in formats:
            try:
                dt = datetime.datetime.strptime(date_string, fmt)
                return dt.date()
            except (TypeError, ValueError, Exception):
                pass

        raise ValueError('Invalid date format {}. Allowed formats are {}'.format(repr(date_string), repr(formats)))

    @classmethod
    def str_date(cls, dt: datetime.date) -> str:
        """Return the date as a string."""
        return dt.strftime(cls.DATE_FORMATS[0])

    def __new__(cls, year, month=0, day=0):
        if isinstance(year, bytes):
            year = year.decode()
        if isinstance(year, (str, datetime.time)):
            dt = cls.make_date(year)
            return dt
            # year = dt.year
            # month = dt.month
            # day = dt.day
        super().__new__(year=year, month=month, day=day)

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
                allowed_formats=cls.DATE_FORMATS,
                )

    @classmethod
    def validate(cls, v):
        if v is None:
            return None
        return Date(v)

    def __str__(self):
        return self.str_date(self)


class Time(datetime.time):
    """DateTime class for pydantic which converts custom string formats into datetimes."""

    TIME_FORMATS = [
        '%I:%M:%S %p',     # '02:24:55 PM'
        '%I:%M:%S.%f %p',  # '02:24:55.000200 PM'
        '%I:%M %p',        # '02:24 PM'
        '%H:%M:%S',        # '14:24:55'
        '%H:%M:%S.%f',     # '14:24:55.000200'
        '%H:%M',           # '14:24'
        ]

    @classmethod
    def make_time(cls, time_string: Union[datetime.time, str], formats: List[str] = None) -> datetime.time:
        """Make the time object from the given time string.

        Args:
            time_string (str): Time string '04:00 PM' ...
            formats (list): List of acceptable time string formats.

        Returns:
            t (datetime.time): Time object or None.
        """
        if isinstance(time_string, datetime.time):
            return time_string

        if formats is None:
            formats = [i for i in cls.TIME_FORMATS]

        for fmt in formats:
            try:
                dt = datetime.datetime.strptime(time_string, fmt)
                return dt.time()
            except (TypeError, ValueError, Exception):
                pass

        raise ValueError('Invalid time format {}. Allowed formats are {}'.format(repr(time_string), repr(formats)))

    @classmethod
    def str_time(cls, dt: datetime.time) -> str:
        """Return the time as a string"""
        return dt.strftime(cls.TIME_FORMATS[0])

    def __new__(cls, hour=0, minute=0, second=0, microsecond=0, tzinfo=None, *, fold=0):
        if isinstance(hour, bytes):
            hour = hour.decode()
        if isinstance(hour, (str, datetime.time)):
            dt = cls.make_time(hour)
            return dt
            # hour = dt.hour
            # minute = dt.minute
            # second = dt.second
            # microsecond = dt.microsecond
            # tzinfo = dt.tzinfo
            # fold = dt.fold
        super().__new__(cls, hour=hour, minute=minute, second=second, microsecond=microsecond, tzinfo=tzinfo, fold=fold)

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
                allowed_formats=cls.TIME_FORMATS,
                )

    @classmethod
    def validate(cls, v):
        if v is None:
            return None
        return Time(v)

    def __str__(self):
        return self.str_time(self)


class DateTime(datetime.datetime):
    """DateTime class for pydantic which converts custom string formats into datetimes."""

    DATETIME_FORMATS = [d + ' ' + t for t in Time.TIME_FORMATS for d in Date.DATE_FORMATS]
    DATETIME_FORMATS.extend(Date.DATE_FORMATS)
    DATETIME_FORMATS.extend(Time.TIME_FORMATS)

    @classmethod
    def make_datetime(cls, date_string: Union[datetime.datetime, str], formats: List[str] = None) -> datetime.datetime:
        """Make the datetime from the given date time string.

        Args:
            date_string (str): Datetime string '04:00 PM' ...
            formats (list): List of acceptable datetime string formats.

        Returns:
            dt (datetime.datetime): Datetime object or None.
        """
        if isinstance(date_string, datetime.datetime):
            return date_string

        if formats is None:
            formats = [i for i in cls.DATETIME_FORMATS]

        for fmt in formats:
            try:
                return datetime.datetime.strptime(date_string, fmt)
            except (TypeError, ValueError, Exception):
                pass

        try:  # Try ISO format
            return datetime.datetime.fromisoformat(date_string)
        except:
            pass

        raise ValueError('Invalid datetime format {}. Allowed formats are {}'.format(repr(date_string), repr(formats)))

    @classmethod
    def str_datetime(cls, dt: datetime.datetime) -> str:
        """Return the datetime as a string."""
        return dt.strftime(cls.DATETIME_FORMATS[0])

    def __new__(cls, year, month=0, day=0, hour=0, minute=0, second=0, microsecond=0, tzinfo=None, *, fold=0):
        if isinstance(year, (str, datetime.datetime)):
            dt = cls.make_datetime(year)
            return dt
            # year = dt.year
            # month = dt.month
            # day = dt.day
            # hour = dt.hour
            # minute = dt.minute
            # second = dt.second
            # microsecond = dt.microsecond
            # tzinfo = dt.tzinfo
            # fold = dt.fold
        return super().__new__(cls, year, month=month, day=day, hour=hour, minute=minute, second=second,
                               microsecond=microsecond, tzinfo=tzinfo, fold=fold)

    @classmethod
    def __get_validators__(cls):
        # one or more validators may be yielded which will be called in the
        # order to validate the input, each validator will receive as an input
        # the value returned from the previous validator
        yield cls.validate

    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(
                allowed_formats=cls.DATETIME_FORMATS,
                )

    @classmethod
    def validate(cls, v):
        if v is None:
            return None
        return DateTime(v)

    def __str__(self):
        return self.str_datetime(self)


class TimeDelta(datetime.timedelta):
    pass
